/***********************************************************************
 *                                                                     *
 * Copyright (c) David L. Mills 1994                                   *
 *                                                                     *
 * Permission to use, copy, modify, and distribute this software and   *
 * its documentation for any purpose and without fee is hereby         *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the copyright notice and this permission       *
 * notice appear in supporting documentation, and that the name        *
 * University of Delaware not be used in advertising or publicity      *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  The University of Delaware makes no      *
 * representations about the suitability this software for any         *
 * purpose. It is provided "as is" without express or implied          *
 * warranty.                                                           *
 *                                                                     *
 ***********************************************************************
 */
/*
 * Program to control ICOM radios
 *
 * Main program
 */
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "icom.h"

#ifndef MSDOS
#include <sys/types.h>
#endif /* MSDOS */

/*
 * Global declarations
 */
int debug = 0;                  /* see icom.h header file */

/*
 * Command decode
 */
static struct cmdtable {
	char name[10];		/* command name */
	int ident;		/* command ident */
	} cmd[] = {
	{"radio", C_RADIO},	/* select radio */
	{"restore", C_INPUT},	/* restore channels */
	{"save", C_OUTPUT},	/* save channels */
	{"chan", C_READ},	/* set/read channel */
	{"display", C_DISP},	/* display freq/mode/offset */
	{"freq", C_FREQ},	/* load frequency */
	{"mode", C_MODE},	/* load mode */
	{"offset", C_SOFFSET},	/* write offset */
	{"write", C_WRITE},	/* write channel */
	{"clear", C_CLEAR},	/* clear channel */
	{"scan", C_SCAN},	/* specify scan frequency range */
	{"go", C_GO},		/* start scan */
	{"up", C_UP},		/* step up */
	{"down", C_DOWN},	/* step down */
	{"debug", C_DEBUG},	/* set debug switch */	
	{"quit", C_QUIT},	/* exit program */
	{"\0", C_ERROR}		/* unknown command */
	};

/*
 * Program variables
 */
FILE *fp_in;                    /* input file */
FILE *fp_out;                   /* output file */
int fd_icom;			/* serial device descriptor */
struct icom *radio;		/* radio structure pointer */
double freq;			/* frequency */
double offset;			/* tx offset */
double first, last, step;	/* scan/step stuff */
int par1, par2, argn;		/* integer parameters */
char *argv[10];			/* command line args */

/*
 * Function prototypes
 */
void display(struct icom *, char *); /* display funk */
int args(struct icom *, char *); /* decode channel argument */

extern struct icom *select_radio(struct icom *, char *);
extern int setchan(struct icom *, int);
extern int readchan(struct icom *);
extern int loadfreq(struct icom *, double);
extern int loadmode(struct icom *, char[]);
extern int loadoffset(struct icom *, double);
extern int writechan(struct icom *);
extern int clearchan(struct icom *);

extern void ascfreq(char *, int *);
extern int doublefreq(double, int *);
extern double freqdouble(int *);
extern void strlwrc(char *);
extern void freqtoa(int *, char *);
extern void modetoa(struct icom *, int , char *);

FILE *fopen();
char *strtok(), *strcpy();

/*
 * Main program
 */
main(argc, argcv)
	int argc;		/* number of command-line arguments */
	char **argcv;		/* vector of command-line argument pointers */

{
	char s1[SMAX], s2[SMAX]; /* parameter strings */
	int i, temp;		/* temps */
	extern char *optarg;
	extern int optind, opterr;

/*
 * Initialize
 */
	radio = NULL;
	init();

#ifndef MSDOS
/*
 * Process command-line arguments
 */
	i = FALSE;
	while ((temp = getopt(argc, argcv, "r:f:m:o:c:d")) != -1) {
		switch (temp) {

		case 'r':
			radio = select_radio(radio, optarg);
			if (radio == NULL) {
				printf("*** unknown radio %s\n",
				    optarg);
				exit(1);
			}
			break;

		case 'm':
			if (radio == NULL) {
				printf("*** radio not specified\n");
				exit(1);
			}
			if (loadmode(radio, optarg)) {
				printf("*** invalid mode\n");
				exit(1);
			}
			i = TRUE;
			break;

		case 'f':
			if (radio == NULL) {
				printf("*** radio not specified\n");
				exit(1);
			}
			sscanf(optarg, "%lf", &freq);
			if (freq > 1000)
				freq /= 1000;
			if (loadfreq(radio, freq)) {
				printf("*** Invalid frequency\n");
				exit(1);
			}
			i = TRUE;
			break;

		case 'o':
			sscanf(optarg, "%lf", &offset);
			if (offset > 1000)
				offset /= 1000;
			if (radio == NULL) {
				printf("*** radio not specified\n");
				exit(1);
			}
 			if (loadoffset(radio, offset)) {
				printf("*** Invalid offset\n");
				exit(1);
			}
			i = TRUE;
			break;

		case 'c':
			sscanf(optarg, "%i", &i);
			if (setchan(radio, i)) {
				printf("*** invalid channel %d\n", i);
				break;
			}
			readchan(radio);
			i = TRUE;
			break;

		case 'd':
			readchan(radio);
			modetoa(radio, radio->mode, s1);
			printf("%.5lf %s", radio->vfo, s1);
			if (radio->flags & F_OFFSET)
				printf(" %.2lf\n", radio->offset);
			else
				printf("\n");
			i = TRUE;
			break;
		}
	}
	if (i == TRUE)
		exit(1);
#endif /* MSDOS */

/*
 * Main loop
 */
	while (1) {

		/*
		 * Get next command and parse tokens. Select command
		 * based on first token.
		 */
		printf("icom>");
		gets(s1, SMAX - 1);
		strlwrc(s1);
		if ((argv[0] = strtok(s1, " ,\n")) == NULL)
			continue;
		for (argn = 1; (argv[argn] = strtok(NULL, " ,\n")) != NULL; argn++);
		for (i = 0; cmd[i].name[0] != '\0'; i++) {
			if (strcmp(cmd[i].name, argv[0]) == 0)
				break;
		}
		switch (cmd[i].ident) {

		/*
		 * radio [name]
		 *
		 * This command selects the named radio for further
		 * commands and displays its description and band
		 * limits. If name is missing, the description and band
		 * limits of the current selection are displayed.
		 */
		case C_RADIO:
			if (argn < 2) {
				if (radio != NULL)
					printf("%s %.1f MHz - %.1f MHz\n",
					    radio->name, radio->lband,
					    radio->uband);
				else
					printf("usage: radio name\n");
				break;
			}
			radio = select_radio(radio, argv[1]);
			if (radio == NULL)
				printf("*** unknown radio %s\n",
				    argv[1]);
			else {
				first = radio->lband;
				last = radio->uband;
				step = .001;
				printf("%s %.1f MHz - %.1f MHz\n",
				    radio->name, radio->lband,
				    radio->uband);
			}
			break;

		/*
		 * chan [first] [last]
		 *
		 * This command displays the memory channels from first
		 * to last. The first and last default to the current
		 * channel; the last defaults to the first, if given.
		 * The last must be at least as great as the first.
		 */
		case C_READ:
			par1 = par2 = radio->mchan;
			if (argn > 1)
				par1 = par2 = args(radio, argv[1]);
			if (argn > 2)
				par2 = args(radio, argv[2]);
			if (par2 < par1)
				par2 = par1;
			for (i = par1; i <= par2; i++) {
				if (setchan(radio, i)) {
					printf("*** invalid channel %d\n",
					    i);
					break;
				}
				if (readchan(radio)) {
					printf("Chan%3i [clear]\n", i);
				} else {
					display(radio, s2);
					printf("Chan%3i %s\n", i, s2);
				}
			}
			break;

		/*
		 * restore file [first] [last]
		 *
		 * This command restores the memory channels from first
		 * to last from the named file. The first defaults to
		 * the current channel; the last defaults to the highest
		 * channel. The last must be at least as great as the
		 * first.
		 */
		case C_INPUT:
			if (argn < 2)
				printf("usage: restore file [first] [last]\n");
			if ((fp_in = fopen(argv[1], "r")) == NULL) {
				printf("*** input file %s not found\n", argv[1]);
				break;
			}
			par1 = radio->mchan;
			par2 = radio->maxch;
			if (argn > 2)
				par1 = args(radio, argv[2]);
			if (argn > 3)
				par2 = args(radio, argv[3]);
			if (par2 < par1)
				par2 = par1;
			for (i = par1; i <= par2; i++) {
				if (fgets(s1, SMAX - 1, fp_in) == NULL)
					break;
				printf("Chan%3i %s", i, s1);
				temp = sscanf(s1, "%lf%s%lf",
				    &freq, s2, &offset);
				if (temp < 2) {
					printf("*** bad file format\n");
					break;
				}
				if (setchan(radio, i)) {
					printf("*** invalid channel\n");
					break;
				}
				if (loadmode(radio, s2)) {
					printf("*** invalid mode\n");
					break;
				}
				if (loadfreq(radio, freq)) {
					printf("*** invalid frequency\n");
					break;
				}
				if ((radio->flags & F_OFFSET) && temp > 2) {
					if (loadoffset(radio, offset)) {
						printf("*** invalid offset\n");
						break;
					}
				}
				if (writechan(radio)) {
					printf("*** invalid write memory\n");
					break;
				}
			}
			close(fp_in);
			break;

		/*
		 * save file [first] [last]
		 *
		 * This command saves the memory channels from first to
		 * last to the named file. The first defaults to the
		 * current channel; the last defaults to the highest
		 * channel. The last must be at least as great as the
		 * first.
		 */
		case C_OUTPUT:
			if (argn < 2)
				printf("usage: save file [first] [last]\n");
			if ((fp_out = fopen(argv[1], "w")) == NULL) {
				printf("*** output file %s error\n", argv[1]);
				break;
			}
			par1 = radio->mchan;
			par2 = radio->maxch;
			if (argn > 2)
				par1 = args(radio, argv[2]);
			if (argn > 3)
				par2 = args(radio, argv[3]);
			if (par2 < par1)
				par2 = par1;
			for (i = par1; i <= par2; i++) {
				if (setchan(radio, i)) {
					printf("*** invalid channel %d\n",
					    i);
					break;
				}
				if (readchan(radio) == 0) {
					display(radio, s2);
					fprintf(fp_out, "%s\n", s2);
				}
			}
			fclose(fp_out);
			break;

		/*
		 * display
		 */
		case C_DISP:
			display(radio, s2);
			printf("%s\n", s2);
			break;

		/*
		 * freq frequency [mode]
		 *
		 * This command sets the current operating frequency and
		 * optional mode. Note that the mode has to be set
		 * first, since some radios shift only the bfo and don't
		 * compensate the local oscillator.
		 *
		 * Watch out for ham/general-coverage switch; these
		 * radios won't let you set the frequency outside the
		 * ham bands if the switch is in the ham position. This
		 * holds also when reading a memory channel, which can
		 * be stored either way.
		 */
		case C_FREQ:
			if (argn > 2) {
				if (loadmode(radio, argv[2])) {
					printf("*** invalid mode\n");
					break;
				}
			}
			if (argn < 2) {
				printf("Frequency %11.5lf\n",
				    radio->vfo);
				break;
			}
			sscanf(argv[1], "%lf", &freq);
			if (freq > 1000)
				freq /= 1000;
			if (loadfreq(radio, freq))
				printf("*** Invalid frequency\n");
			break;

		/*
		 * mode mode
		 *
		 * This command sets the current operating mode. The
		 * mode names are specific to each radio type; not all
		 * radios support the same modes. Note that the
		 * frequency has to be reset, since some radios shift
		 * only the bfo and don't compensate the local
		 * oscillator.
		 */
		case C_MODE:
			if (argn < 2) {
				modetoa(radio, radio->mode, s1);
				printf("Mode %s\n", s1);
				break;
			}
			if (loadmode(radio, argv[1])) {
				printf("*** invalid mode\n");
				break;
			}
			if (loadfreq(radio, radio->vfo))
				printf("*** Invalid frequency\n");
			break;

		/*
		 * offset offset
		 *
		 * This command selects the current tx offset for vhf
		 * transceivers. It is not clear how to set the +-duplex
		 * switch.
		 */
		case C_SOFFSET:
			if (argn < 2) {
				printf("TX offset%5.2lf\n",
				    radio->offset);
				break;
			}
			sscanf(argv[1], "%lf", &offset);
			if (loadoffset(radio, offset))
				printf("*** invalid offset\n");
			break;

		/*
		 * write [chan]
		 *
		 * This command writes the current frequency, mode and
		 * offset to the selected memory channel, with default
		 * the current channel.
		 */
		case C_WRITE:
			temp = radio->mchan;
			if (argn > 1)
				if (setchan(radio, args(radio,
				    argv[1]))) {
					printf("*** invalid channel\n");
					break;
				}
			if (writechan(radio))
				printf("*** invalid write memory\n");
			if (argn > 1)
				setchan(radio, temp);
			break;

		/*
		 * clear [chan]
		 *
		 * This command clears the selected memory channel, with
		 * default the current channel. Currently, it works only
		 * with the r7000.
		 */
		case C_CLEAR:
			temp = radio->mchan;
			if (argn > 1)
				if (setchan(radio, args(radio,
				    argv[1]))) {
					printf("*** invalid channel\n");
					break;
				}
			if (clearchan(radio))
				printf("*** invalid clear memory\n");
			if (argn > 1)
				setchan(radio, temp);
			break;

		/*
		 * scan first last step
		 */
		case C_SCAN:
			if (argn == 1) {
				printf("first %.5lf last %.5lf step %.5lf\n",
				    first, last, step);
				break;
			}
			if (argn < 4) {
				printf("usage: scan first last step\n");
				break;
			}
			sscanf(argv[1], "%lf", &first);
			sscanf(argv[2], "%lf", &last);
			sscanf(argv[3], "%lf", &step);
			break;

		/*
		 * go
		 */
		case C_GO:
			for (freq = first; freq <= last; freq += step)
				if (loadfreq(radio, freq)) {
					printf("*** invalid frequency\n");
				}
			break;

		/*
		 * up
		 */
		case C_UP:
			freq = radio->vfo + step;
			if (freq > last)
				freq = first;
			if (loadfreq(radio, freq))
				printf("*** invalid frequency\n");
			else {
				display(radio, s1);
				printf("%s\n", s1);
			}
			break;

		/*
		 * down
		 */
		case C_DOWN:
			freq = radio->vfo - step;
			if (freq < first)
				freq = last;
			if (loadfreq(radio, freq))
				printf("*** invalid frequency\n");
			else {
				display(radio, s1);
				printf("%s\n", s1);
			}
			break;

		/*
		 * quit
		 */
		case C_QUIT:
			exit(0);

		/*
		 * debug bits
		 */
		case C_DEBUG:
			if (argn < 2)
				debug = 0;
			else
				sscanf(argv[1], "%x", &debug);
			break;

		/*
		 * unknown command
		 */
		default:
			printf("*** unknown command %s\n", argv[0]);
		}
	}
}

/*
 * display(radio, sptr) - display frequency, mode and tx offset
 */
void display(radio, sptr)
	struct icom *radio;	/* radio structure */
	char *sptr;		/* ascii output string */
{
	char s1[SMAX];

	modetoa(radio, radio->mode, s1);
	if (radio->flags & F_OFFSET)
		sprintf(sptr, "%10.5lf %s%5.2lf",
			radio->vfo, s1, radio->offset);
	else
		sprintf(sptr, "%10.5lf %s", radio->vfo, s1);
}

/*
 * args(radio, sptr) - decode channel argument
 *
 * .	current channel
 * *	first channel (default)
 * $	last channel
 * dig	specified channel
 *
 * argument is clamped at low and high limits of valid range.
 */
int args(radio, sptr)		/* returns 0 (ok), 1 (error) */
	struct icom *radio;	/* radio structure */
	char *sptr;		/* ascii argument pointer */
{
	int temp;

	temp = 0;
	if (*sptr == '.')
		temp = radio->mchan;
	if (*sptr == '$')
		temp = radio->maxch;
	if (isdigit(*sptr))
		temp = atoi(sptr);
	if (temp < 0)
		temp = 0;
	if (temp > radio->maxch)
		temp = radio->maxch;
	if (temp == 0 && (radio->flags & F_VFO))
		temp = 1;
	return(temp);
}

/* end program */
