/***********************************************************************
 *                                                                     *
 * Copyright (c) David L. Mills 1994                                   *
 *                                                                     *
 * Permission to use, copy, modify, and distribute this software and   *
 * its documentation for any purpose and without fee is hereby         *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the copyright notice and this permission       *
 * notice appear in supporting documentation, and that the name        *
 * University of Delaware not be used in advertising or publicity      *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  The University of Delaware makes no      *
 * representations about the suitability this software for any         *
 * purpose. It is provided "as is" without express or implied          *
 * warranty.                                                           *
 *                                                                     *
 ***********************************************************************
 */
/*
 * Program to control ICOM radios
 *
 * Input/output interface
 */
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "icom.h"

#ifndef MSDOS
#include <sys/fcntl.h>
#include <sys/termios.h>
#include <sys/unistd.h>
#else /* MSDOS */
#include <string.h>
#endif /* MSDOS */

/*
 * Parameters
 */
#define PR 0xFE			/* preamble */
#define TX 0xE0			/* controller address */
#define FI 0xFD			/* end of message */
#define ACK 0xFB		/* CN normal reply */
#define NAK 0xFA		/* CN error reply */

#define TXRETRY 3		/* retransmission counter */
#define XFRETRY 3000		/* interface timeout counter (MSDOS) */

#ifdef MSDOS
/*
 * Define port and speed
 */
#define PORT 0x03f8		/* port address (COM1) */
/* #define PORT 0x02f8		/* port address (COM2) */
/* #define PORT 0x03e8		/* port address (COM3) */
/* #define PORT 0x02e8		/* port address (COM4) */
/* #define BAUD 384		/* baud rate 300 */
#define BAUD 96			/* baud rate 1200 */
/* #define BAUD 12		/* baud rate 9600 */

/*
 * Serial port definitions (8250)
 */
#define THR 0			/* transmitter holding register */
#define RBR 0			/* receiver buffer register */
#define DLL 0			/* divisor latch LSB */
#define DLM 1			/* divisor latch MSB */
#define LCR 3			/* line control register */
#define LCR_8BITS 3		/* 8 bit words */
#define   LCR_DLAB 0x80		/* divisor latch access bit */
#define MCR 4			/* modem control register */
#define   MCR_DTR 1		/* data terminal ready */
#define   MCR_RTS 2		/* request to send */
#define LSR 5			/* line status register */
#define   LSR_DR 0x01		/* data ready */
#define   LSR_THRE 0x20		/* trans line holding register empty */
#define   LSR_TSRE 0x40		/* transmitter shift register empty */
#else /* MSDOS */
#define DICOM "/dev/icom"	/* serial line driver */
#define	BICOM B1200		/* serial line speed */
#endif /* MSDOS */

/*
 * Global declarations
 */
extern int debug;		/* debug flags */

/*
 * Local declarations
 */
int count;			/* retry counter */
int fd_icom;			/* radio file descriptor */

/*
 * Function prototypes
 */
int sndoctet(int);		/* send octet */
int rcvoctet(void);		/* receive octet */

#ifdef MSDOS
int inp(int);			/* input port byte */
void outp(int, int);		/* output port byte */
void outpw(int, int);		/* output port word */
#endif /* MSDOS */

/*
 * Packet routines
 *
 * These routines send a packet and receive the response. If a
 * collision is detected on transmit (by the interface routines) the
 * packet is resent. If a preamble error occurs on receive the response
 * is discarded (including the terminating FI) and the packet is
 * resent. If the maximum number of retries is exceeded, the program
 * aborts.
 *
 * ICOM frame format
 *
 * Frames begin with a two-octet preamble PR-PR followed be the
 * transceiver address RE, controller address TX, control code CN, a
 * number of data octets DA (depending on command) and the terminator
 * FI. Since the bus is bidirectional, every octet output by the
 * computer is echoed on input. Every valid frame sent by the computer
 * is answered with a frame in the same format, but with the RE and TX
 * fields interchanged. The CN field is set to NAK in case of error and
 * to ACK in case of a valid command that returns no data.
 *
 *	+------+------+------+------+------+--//--+------+
 *	|      |      |      |      |      |      |      |
 *	|  PR  |  PR  |  RE  |  TX  |  CN  |  DA  |  FI  |
 *	|      |      |      |      |      |      |      |
 *	+------+------+------+------+------+--//--+------+
 */

/*
 * initpkt() - initialize serial interface
 *
 * This routine opens the serial interface for raw transmission; that
 * is, character-at-a-time, no stripping, checking or monkeying with the
 * bits. For Unix, an input operation ends either with the receipt of a
 * character or a one-second timeout.
 */
void initpkt()
{
#ifdef MSDOS
	outp(PORT+LCR, LCR_DLAB); /* set baud */
	outpw(PORT+DLL, BAUD);
	outp(PORT+LCR, LCR_8BITS); /* set 8 bits, no parity */
	outp(PORT+MCR, MCR_DTR+MCR_RTS); /* wake up modem */
#else /* MSDOS */
	struct termios ttyb, *ttyp;

	ttyp = &ttyb;
	if ((fd_icom = open(DICOM, O_RDWR, 0777)) < 0) {
		printf("\n*** Unable to open serial line %s\n", DICOM);
		exit(1);
	}
	if (tcgetattr(fd_icom, ttyp) < 0) {
		printf("\n*** Unable to get line parameters\n");
		exit(1);
	}
	ttyp->c_iflag = 0;	/* input modes */
	ttyp->c_oflag = 0;	/* output modes */
	ttyp->c_cflag = BICOM|CS8|CREAD|CLOCAL;	/* control modes */
	ttyp->c_lflag = 0;	/* local modes */
	ttyp->c_cc[VMIN] = 0;	/* min chars */
	ttyp->c_cc[VTIME] = 10;	/* receive timeout */
	if (tcsetattr(fd_icom, TCSANOW, ttyp) < 0) {
		printf("\n*** Unable to set line parameters\n");
		exit(1);
	}
#endif /* MSDOS */
}
/*
 * sndpkt(r, x, y) - send packet and receive response
 *
 * This routine sends a command vector, which consists of all except the
 * preamble octets PR-PR. It then listens for the response vector and
 * returns all except the preamble octets to the caller. All octets sent
 * are echo checked; if any mismatch, the operation is retried up to
 * TXRETRY times. The routine checks for correct response header format;
 * that is, the length of the response vector returned to the caller
 * must be at least 2 and the RE and TX octets must be interchanged;
 * otherwise, the operation is retried.
 *
 * The trace function, which is enabled by the X_TRACE bit in the debug
 * word, prints all characters received or echoed on the bus preceded by
 * a T (transmit) or R (receive). 
 */
int sndpkt(r, x, y)		/* returns count */
	int r;			/* radio address */
	int x[];		/* command vector */
	int y[];		/* response vector */
{
	int i, j, k, temp;	/* temps */

#ifndef MSDOS
	(void)tcflush(fd_icom, TCIOFLUSH);
#endif /* MSDOS */
	for (i = 0; i < TXRETRY; i++) {

		/*
		 * Transmit packet. If echo check fails, try again.
		 */
		if (debug & X_TRACE)
			printf("T:");
		if (sndoctet(PR) != PR) /* send header */
			continue;
		if (sndoctet(PR) != PR)
			continue;
		if (sndoctet(r) != r)
			continue;
		if (sndoctet(TX) != TX)
			continue;
		for (j = 0; (j < BMAX); j++) /* send body */
			if (sndoctet(x[j]) == FI)
				break;
		if (x[0] <= 1)
			return(0); /* shortcut for broadcast */

		/*
		 * Receive packet. First, delete all characters
		 * preceeding a PR, then discard all PRs. Check that the
		 * RE and TX fields are correctly interchanged, then
		 * copy the remaining data and FI to the user buffer.
		 */
		if (debug & X_TRACE)
			printf("R:");
		k = 0;
		while ((temp = rcvoctet()) != FI) {
			if (k > 0)
				continue;
			if (temp != PR) /* header */
				continue;
			if (rcvoctet() != PR)
				continue;
			if (rcvoctet() != TX) {
				k++;
				if (debug & X_ERMSG)
					printf("\n*** TX error\n");
				continue;
			}
			if (rcvoctet() != r) {
				k++;
				if (debug & X_ERMSG)
					printf("\n*** RE error\n");
				continue;
			}
			for (j = 0; (j < BMAX); j++) { /* body */
				if ((y[j] = rcvoctet()) == FI) {
					if (j < 1) {
						if (debug & X_ERMSG)
							printf("\n*** runt packet\n");
						continue;
					}
					return(j);
				}
			}
		}
	}
	if (debug & X_ERMSG)
		printf("\n*** packet error\n");
	return(0);
}

/*
 * Interface routines
 *
 * These routines read and write octets on the bus. In case of receive
 * timeout a FI code is returned. In case of output collision (echo
 * does not match octet sent), the remainder of the collision frame
 * (including the trailing FI) is discarded.
 */

#ifdef MSDOS
/*
 * sndoctet(x) - send octet
 */
int sndoctet(x)			/* returns octet */
	int x;			/* octet */
{
	int temp;

	inp(PORT+RBR);		/* flush spikes */
	outp(PORT+THR, x);	/* write octet */
	if (x == (temp = rcvoctet()))
		return(temp);	/* return echo */
	if (debug & X_ERMSG)
		printf("\n*** bus error %02x %02x\n", x, temp);
	while (rcvoctet() != FI); /* discard junk */
	return(FI);
}

/*
 * Receive octet
 *     returns octet received, force FI if error
 */
int rcvoctet()		/* returns octet */
{
	int temp, i;

	for (i = 0; i < XFRETRY; i++) /* wait for input */
	if ((inp(PORT+LSR))&LSR_DR != 0) {
		if (i > count)
			count = i;
		temp = inp(PORT+RBR); /* read octet */
		return(temp);
	}
	return(FI);
}
#else /* MSDOS */
/*
 * sndoctet(x) - send octet
 */
int sndoctet(x)		/* returns octet */
	int x;		/* octet */
{
	unsigned char temp[1];
	int word;

	temp[0] = x;
	write(fd_icom, temp, 1); /* write octet */
	word = rcvoctet();
	if (x == word)
		return(x);	/* return echo */
	if (debug & X_ERMSG)	/* come here if collision */
		printf("\n*** bus error %02x %02x\n", x, word);
	while (rcvoctet() != FI); /* discard junk */
	return(FI);
}

/*
 * rcvoctet () - receive octet
 */
int rcvoctet()		/* returns octet */
{
	unsigned char temp[1];

	if (read(fd_icom, temp, 1) < 1) /* read octet */
		temp[0] = FI;	/* come here if timeout */
	if (debug & X_TRACE) {
		printf(" %02x", temp[0]);
		if (temp[0] == FI)
			printf("\n");
	}
	return(temp[0]);
}
#endif /* MSDOS */

/* end program */
