/***********************************************************************
 *                                                                     *
 * Copyright (c) David L. Mills 1994                                   *
 *                                                                     *
 * Permission to use, copy, modify, and distribute this software and   *
 * its documentation for any purpose and without fee is hereby         *
 * granted, provided that the above copyright notice appears in all    *
 * copies and that both the copyright notice and this permission       *
 * notice appear in supporting documentation, and that the name        *
 * University of Delaware not be used in advertising or publicity      *
 * pertaining to distribution of the software without specific,        *
 * written prior permission.  The University of Delaware makes no      *
 * representations about the suitability this software for any         *
 * purpose. It is provided "as is" without express or implied          *
 * warranty.                                                           *
 *                                                                     *
 ***********************************************************************
 */
/*
 * Program to control ICOM radios
 * 
 * Subroutine library
 */
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "icom.h"

#ifndef MSDOS
#include <fcntl.h>
#include <sys/types.h>
#include <sys/termios.h>
#endif /* MSDOS */

/*
 * Parameters
 */
#define PR 0xFE			/* preamble */
#define TX 0xE0			/* controller address */
#define FI 0xFD			/* end of message */
#define ACK 0xFB		/* CN normal reply */
#define NAK 0xFA		/* CN error reply */

/*
 * HF/VHF/UHF transceiver and receiver mode decode
 *
 * Note that not all radios speak all modes.
 */
struct modestruct mode[] = {
	{"LSB", 0x0000},	/* LSB */
	{"LSB", 0x8100},	/* LSB (R72) */
	{"USB", 0x0001},	/* USB */
	{"USB", 0x8101},	/* USB (R72) */
	{"AM", 0x0002},		/* AM */
	{"AM", 0x8102},		/* AM (R72) */
	{"AMn", 0x8202},	/* AM narrow (R72) */
	{"CW", 0x0003},		/* CW */
	{"CW", 0x8103},		/* CW (R72) */
	{"CWn", 0x8203},	/* CW narrow */
	{"RTTY", 0x0004},	/* RTTY */
	{"RTTYn", 0x8204},	/* RTTY narrow */
	{"FM", 0x0005},		/* FM */
	{"FM", 0x8105},		/* FM (R72) */
	{"\0", 0x00ff}		/* end of table */
	};

/*
 * VHF/UHF receiver mode decode (r7000)
 *
 * Note that panel FM mode is really wideband FM and panel FMn mode is
 * really FM mode on other radios. Also, panel SSB mode is mapped to
 * USB.
 */
struct modestruct spmode[] = {
	{"AM", 0x0002},		/* AM */
	{"FMw", 0x0005},	/* FM wide */
	{"FM", 0x8205},		/* FM */
	{"USB", 0x8005},	/* USB */
	{"?", 0x00ff},		/* clear channel */
	{"\0", 0x00ff}		/* end of table */
	};

/*
 * Name decode
 */
struct namestruct {
	char name[30];		/* radio name */
	int ident;		/* bus address */
	int maxch;		/* max memory channel */
	struct modestruct *mode; /* mode table */
	int flags;		/* flag bits */
	} name[] = {
	{"IC-271 VHF Transceiver", 0x20, 32, mode, F_MHZ|F_OFFSET},
	{"IC-275 VHF Transceiver", 0x10, 99, mode,
	    F_MHZ|F_OFFSET|F_SCAN},
	{"IC-375 VHF Transceiver", 0x12, 99, mode,
	    F_MHZ|F_OFFSET|F_SCAN},
	{"IC-471 UHF Transceiver", 0x22, 32, mode, F_MHZ|F_OFFSET},
	{"IC-475 UHF Transceiver", 0x14, 99, mode,
	    F_MHZ|F_OFFSET|F_SCAN},
	{"IC-575 VHF Transceiver", 0x16, 99, mode,
	    F_MHZ|F_OFFSET|F_SCAN},
	{"IC-725 HF Transceiver", 0x04, 26, mode, 0},
	{"IC-735 HF Transceiver", 0x28, 12, mode, 0},
	{"IC-751 HF Transceiver", 0x1c, 32, mode, 0},
	{"IC-761 HF Transceiver", 0x1e, 32, mode, 0},
	{"IC-781 HF Transceiver", 0x26, 32, mode, 0},
	{"IC-1271 UHF Transceiver", 0x24, 32, mode, F_MHZ|F_OFFSET},
	{"IC-R71A HF Receiver", 0x1A, 32, mode, F_RCV},
	{"IC-R72A HF Receiver", 0x32, 99, mode, F_RCV},
	{"IC-R7000 VHF/UHF Receiver", 0x08, 99, spmode,
	    F_RCV|F_VFO|F_MHZ},
	{"\0", 0x0, 0, NULL, 0}	/* end of table */
	} ;

/*
 * Radio command prototypes
 */
int s_freqt[] = {0x00,0,0,0,0,0,FI}; /* freq set (transceive mode) */
int s_modet[] = {0x01,0,FI};	/* set mode (transceive) */
int s_rband[] = {0x02,FI};	/* read band edge */
int s_rfreq[] = {0x03,FI};	/* read frequency */
int s_rmode[] = {0x04,FI};	/* read mode */
int s_sfreq[] = {0x05,0,0,0,0,0,FI}; /* set frequency */
int s_smode1[] = {0x06,0,FI};	/* set mode */
int s_smode2[] = {0x06,0,0,FI};	/* set extended mode (r7000 only) */
int s_svfo[] = {0x07,FI};	/* set vfo (except r7000) */
int s_chan[] = {0x08,0,FI};	/* set memory channel */
int s_write[] = {0x09,FI};	/* memory write */
int s_rvfo[] = {0x0a,FI};	/* vfo write (memory -> vfo)
				 * (except r7000) */
int s_clear[] = {0x0b,FI};	/* memory clear (275/375/575/r7000
				 * only) */
int s_roffset[] = {0x0c,FI};	/* read tx offset frequency
				 * (VHF/UHF transceivers only) */
int s_soffset[] = {0x0d,0,0,0,0,0,FI}; /* set tx offset frequency
				 * (VHF/UHF transceivers only) */
int s_scan[] = {0x0e,0,FI};	/* scan control (275/375/575 only) */

/*
 * Function prototypes
 */
struct icom *select_radio(struct icom *, char *); /* select radio */
int setchan(struct icom *, int); /* set radio channel */
int readchan(struct icom *);	/* read frequency and mode */
int loadfreq(struct icom *, double); /* load frequency */
int loadmode(struct icom *, char[]); /* load mode */
int loadoffset(struct icom *, double); /* write offset */
int writechan(struct icom *);	/* write memory channel */
int clearchan(struct icom *);	/* clear memory channel */

void ascfreq(char *, int *);	/* convert ascii to frequency */
void doublefreq(double, int *);	/* convert long to frequency */
double freqdouble(int *);	/* convert frequency to long */
void strlwrc(char *);		/* convert string to lower case */
void freqtoa(int *, char *);	/* convert frequency to ascii */
void modetoa(struct icom *, int , char *); /* convert mode to ascii */
int atomode(struct icom *, char *); /* convert ascii to mode */

void initpkt(void);		/* initialize packet interface */
int sndpkt(int, int[], int[]);	/* send packet and receive reply */

/*
 * Initialize
 */
void init() {
	initpkt();		/* initialize (device dependent) */
	}

/*
 * select_radio(rad, ptr) - select/initalize radio structure
 */
struct icom *select_radio(rad, ptr) /* returns structure pointer (ok)
				 * or NULL (error) */
	struct icom *rad;	/* radio structure pointer */
	char *ptr;		/* radio name string */
{
	struct icom *radio;
	double temp1;
	int buf[BMAX], temp, i;
	char s2[SMAX];

	/*
	 * Free existing structure (if present).
	 */
	if (rad != NULL)
		free(rad);

	/*
	 * Find prototype; return NULL if not found.
	 */
	for (i = 0; name[i].name[0] != '\0'; i++) {
		strcpy(s2, name[i].name);
		strlwrc(s2);
		if (strstr(s2, ptr) != NULL)
			break;
	}
	if (name[i].name[0] == '\0')
		return(NULL);

	/*
	 * Found. Initialize structure and read band edges.
	 */
	radio = (struct icom *)malloc(sizeof(struct icom));
	strcpy(radio->name, name[i].name);
	radio->ident = name[i].ident;
	radio->maxch = name[i].maxch;
	radio->flags = name[i].flags;
	radio->mchan = 0;
	radio->mode = 0;
	radio->modetab = name[i].mode;
	radio->vfo = 0xff;
	radio->offset = 0;
	temp = sndpkt(radio->ident,s_rband, buf);
	if (temp < 12 || buf[0] != s_rband[0]) {
		free(radio);
		return(NULL);
	}
	radio->lband = freqdouble(&buf[1]) / 1e6;
	radio->uband = freqdouble(&buf[7]) / 1e6;
	if (radio->lband > radio->uband) {
		temp1 = radio->lband;
		radio->lband = radio->uband;
		radio->uband = temp1;
	}
	return(radio);
}

/*
 * setchan(radio, chan) - set memory channel number
 */
int setchan(radio, chan)	/* returns 0 (ok), 1 (error) */
	struct icom *radio;	/* radio structure */
	int chan;		/* channel number */
{
	int buf[BMAX], temp;

	if (chan > radio->maxch || chan < 0)
		return(1);
	if (chan == 0 && (radio->flags & F_VFO) == 0) {
		temp = sndpkt(radio->ident, s_svfo, buf);
		
	} else {
		s_chan[1] = chan % 10 + chan / 10 * 16;
		temp = sndpkt(radio->ident, s_chan, buf);
	}
	if (temp < 1 || buf[0] != ACK)
		return(1);
	radio->mchan = chan;
	return(0);
}

/*
 * readchan(radio, result) - read current channel, frequency, mode
 *	and offset (vhf transceivers only)
 */
int readchan(radio)		/* returns 0 (ok), 1 (error) */
	struct icom *radio;	/* radio structure */
{
	int buf[BMAX], temp, i;

	radio->mode = 0xff;
	radio->vfo = 0;
	radio->offset = 0;

	/*
	 * Read mode
	 */
	temp = sndpkt(radio->ident, s_rmode, buf);
	if (temp < 2 || buf[0] != s_rmode[0])
		return(1);
	radio->mode = buf[1];
	if (temp > 2)
		radio->mode = radio->mode | (buf[2] << 8) | 0x8000;

	/*
	 * Read frequency
	 */
	temp = sndpkt(radio->ident, s_rfreq, buf);
	if (temp < 6 || buf[0] != s_rfreq[0])
		return(1);
	radio->vfo = freqdouble(&buf[1]) / 1e6;

	/*
	 * Read offset (vhf transceivers only)
	 */
	if ((radio->flags & F_OFFSET) == 0)
		return(0);
	temp = sndpkt(radio->ident, s_roffset, buf);
	if (temp < 4 || buf[0] != s_roffset[0])
		return(1);
	for (i = temp; i < 6; i++)
		buf[i] = 0;
	radio->offset = freqdouble(&buf[1]) / 1e4;
	return(0);
}

/*
 * loadfreq(radio, freq) - load frequency
 */
int loadfreq(radio, freq)	/* returns 0 (ok), 1 (error) */
	struct icom *radio;	/* radio structure */
	double freq;		/* frequency (MHz) */
{
	int buf[BMAX], temp;
	char s1[SMAX];

	if (freq < radio->lband || freq > radio->uband)
		return(1);
	sprintf(s1, "%10.0lf", freq * 1e6);
	ascfreq(s1, &s_sfreq[1]);
	temp = sndpkt(radio->ident, s_sfreq, buf);
	if (temp < 1 || buf[0] != ACK)
		return(1);
	radio->vfo = freq;
	return(0);
	}

/*
 * loadmode(radio, mode) - load mode
 */
int loadmode(radio, s)		/* returns 0 (ok), 1 (error) */
	struct icom *radio;	/* radio structure */
	char s[];		/* ascii mode string */
{
	int buf[BMAX], temp, mode;

	mode = atomode(radio, s);
	if (mode == 0x00ff)
		return(1);
	if ((mode & 0x8000) == 0) {
		s_smode1[1] = mode;
		temp = sndpkt(radio->ident, s_smode1, buf);
	} else {
		s_smode2[1] = mode & 0xff;
		s_smode2[2] = (mode >> 8) & 0x7f;
		temp = sndpkt(radio->ident, s_smode2, buf);
	}
	if (temp < 1 || buf[0] != ACK)
		return(1);
	radio->mode = mode;
	return(0);
}

/*
 * loadoffset(radio, offset) - load tx offset (vhf transceivers only)
 */
int loadoffset(radio, freq)	/* returns 0 (ok), 1 (error) */
	struct icom *radio;	/* radio structure */
	double freq;		/* tx offset (MHz) */
{
	int buf[BMAX], temp;
	char s1[SMAX];

	sprintf(s1, "%10.0lf", freq * 1e4);
	ascfreq(s1, &s_soffset[1]);
	s_soffset[4] = FI;
	temp = sndpkt(radio->ident, s_soffset, buf);
	if (temp < 1 || buf[0] != ACK)
		return(1);
	radio->offset = freq;
	return(0);
}

/*
 * writechan(radio) - write the current memory channel
 */
int writechan(radio)		/* returns 0 (ok), 1 (error) */
	struct icom *radio;	/* radio structure */
{
	int buf[BMAX], temp;

	temp = sndpkt(radio->ident, s_write, buf);
	if (temp < 1 || buf[0] != ACK)
		return(1);
	return(0);
	}

/*
 * clearchan(radio) - clear the current memory channel
 */
int clearchan(radio)		/* returns 0 (ok), 1 (error) */
	struct icom *radio;	/* radio structure */
{
	int buf[BMAX], temp;

	temp = sndpkt(radio->ident, s_clear, buf);
	if (temp < 1 || buf[0] != ACK)
		return(1);
	radio->mode = 0xff;
	radio->vfo = 0;
	radio->offset = 0;
	return(0);
	}

/*
 * Conversion routines
 *
 * These routines convert between internal formats and radio format.
 * Memory-channel and frequency data are in 4-bit BCD digits, packed
 * two digits per octet. Memory-channel data consist of two digits.
 * Frequency data consist of ten digits in reverse order, in units of
 * Hz for HF radios and kHz for VHF/UHF radios.
 */
/*
 * modetoa(radio, mode, sptr) - convert mode to ascii string
 */
void modetoa(radio, mode, sptr)
	struct icom *radio;	/* radio structure */
	int mode;		/* mode code */
	char *sptr;		/* ascii mode string */
{
	int i;

	for (i = 0; radio->modetab[i].name[0] != '\0'; i++)
		if (radio->modetab[i].mode == mode) {
			sprintf(sptr, "%s", radio->modetab[i].name);
			return;
		}
	sprintf(sptr, "%2x", mode);
}
/*
 * atomode(radio, sptr) - convert ascii string to mode
 */
int atomode(radio, sptr)	/* returns mode (ok), 0x00ff (error) */
	struct icom *radio;	/* radio structure */
	char *sptr;		/* ascii mode string */
{
	int i;
	char s1[SMAX], s2[SMAX];

	strcpy(s1, sptr);
	strlwrc(s1);
	for (i = 0; radio->modetab[i].name[0] != '\0'; i++) {
		strcpy(s2, radio->modetab[i].name);
		strlwrc(s2);
		if (strcmp(s1, s2) == 0)
			break;
	}
	return(radio->modetab[i].mode);
}

/*
 * ascfreq(asciip, rfp) - convery ascii frequency to radio frequency
 */
void ascfreq(x, y)
	char *x; int *y;
{
	int i, temp;

	y = y + 4;
	for (i = 0; i < 5; i++) {	/* pack and reverse */
		temp = (*x &0x0f)<<4; x++;
		*y = temp | (*x &0x0f);
		x++; y--;
	}
}

/*
 * doublefreq(freq, rfp) - convert double to radio frequency
 */
void doublefreq(x, y)
	double x;		/* double frequency */
	int *y;			/* radio frequency */
{
	char s[11];

	sprintf(s, "%10lf", x);	/* convert to ASCII */
	ascfreq(s, y);		/* convert to icom */
}

/*
 * freq = freqdouble(rfp) - convert radio frequency to double
 */
double freqdouble(x)		/* returns frequency */
	int *x;			/* radio frequency */
{
	int i;
	char s[11];
	char *y;
	double temp2;

	temp2 = 0;
	if (*x == 0xff)
		return(temp2);
	x = x + 4;
	y = s;
	for (i = 0; i < 5; i++) { /* unpack and reverse */
		*y = (char)(((*x >> 4) & 0x0f) + '0');
		y++;
		*y = (char)((*x & 0x0f)+'0');
		y++; x--;
		*y = '\0';	/* store terminator */
	}
	sscanf(s, "%lf", &temp2);
	return(temp2);
}

/*
 * strlwrc(string) - convert string to lower case
 */
void strlwrc(s)
	char *s;
{
	char *p;

	for (p = s; *p != '\0'; p++)
		if (isupper(*p))
			*p = tolower(*p);
}

/* end program */
